# coding: utf-8

from __future__ import print_function

import gzip
import os
import re
import shutil
import sys
from collections import OrderedDict

import lz4.frame


def _log(msg, level="info"):
    print("[{level}]{msg}".format(level=level.upper(), msg=msg))


def restore_wal(src_file, target_path):
    if not os.path.exists(target_path):
        os.mkdir(target_path)
    backup_list_name = "BACKUP_INDEX.LST"
    backup_list_file = os.path.join(target_path, backup_list_name)
    tmp_buffer_size = 16384
    with open(src_file, 'rb') as fd, open(backup_list_file, 'wb') as fd_index:
        ch = b'0'
        offset = 0
        index_str = bytes()
        while ch != '\n'.encode():
            ch = fd.read(1)
            index_str += ch
            offset += 1
            fd_index.write(ch)

        pattern = r'(\S+) (\d+)'.encode()
        matcher = re.match(pattern, index_str)
        if not matcher:
            _log("parse index file record failed", "error")
            raise Exception("parse index file record failed")
        if matcher.group(1) != backup_list_name.encode():
            _log("index file name not matched", "error")
            raise Exception("index file name not matched")
        file_size = int(matcher.group(2))

        file_lst = OrderedDict()
        got_bytes = offset
        new_str = bytes()
        while got_bytes < file_size:
            ch = fd.read(1)
            if not ch:
                _log("reached end of stream")
                break
            got_bytes += 1
            new_str += ch
            fd_index.write(ch)

            if ch == '\n'.encode():
                matcher = re.match(pattern, new_str)
                if not matcher:
                    _log("parse index file record failed", "error")
                    raise Exception("parse index file record failed")
                file_lst[matcher.group(1).decode()] = int(matcher.group(2))
                new_str = bytes()
        fd_index.close()

        for name, size in file_lst.items():
            output_file = os.path.join(target_path, name)
            with open(output_file, 'wb') as out_f:
                got_bytes = 0
                while size > got_bytes:
                    if size - got_bytes > tmp_buffer_size:
                        read_len = tmp_buffer_size
                    else:
                        read_len = size - got_bytes
                    out_f.write(fd.read(read_len))
                    got_bytes += read_len
            if re.match(r"[A-F0-9]{24}\.gz", name):
                _decompress_wal(output_file)
        fd.close()

    if os.path.exists(backup_list_file):
        os.remove(backup_list_file)


def _decompress_wal(filename):
    filename = filename.encode()
    try:
        with lz4.frame.open(filename, "rb") as r_fd, open(filename[:-3], "wb") as w_fd:
            w_fd.write(r_fd.read())
            w_fd.flush()
            os.fsync(w_fd.fileno())
    except Exception:
        try:
            with gzip.open(filename, "rb") as r_fd, open(filename[:-3], "wb") as w_fd:
                shutil.copyfileobj(r_fd, w_fd)
        except Exception as err_msg:
            _log("Decompress xlog(%s) failed. Error Msg: %s" % (filename, err_msg), "error")
            return
    os.remove(filename)


def print_help():
    help_msg = """usage: python restore_wal.py src_file_dir target_dir.
    src_file_dir: directory where incremental backup files download from obs locate.
    target_dir: target directory to save pg wal files.
    """
    print(help_msg)


if __name__ == "__main__":
    if len(sys.argv) < 3:
        print_help()
        exit(0)
    wal_dir = sys.argv[1]
    for wal in os.listdir(wal_dir):
        if wal.endswith(".tar"):
            restore_wal(os.path.join(wal_dir, wal), sys.argv[2])
